//_ adi.d
// Copyright (c) 2000-2002 by Digital Mars
// All Rights Reserved
// www.digitalmars.com
// Written by Walter Bright

// Dynamic array property support routines

module internal.adi;

//debug=adi;        // uncomment to turn on debugging printf's

import std.c.stdio;
import std.c.stdlib;
import std.string;
import std.outofmemory;

import internal.arraycat;


extern (C)
bit _darray_in (Array array, TypeInfo type, ...)
{
    void *data = (void *) (&type + 1);
    int size = array.length ();
    int each = type.tsize ();
    
    for (int c; c < size; c ++)
    {
        if (type.equals (array.data + each * c, data))
            return true;
    }
    
    return false;
}

extern (C) Array _darray_new_bit (int length)
{
    Array r;
    int size;
    
    size = (length + 31) / 32;
    r.data = (void *) new uint [size];
    r.flength = length;
    return r;
}

extern (C) bit _darray_get_bit (int index, byte *data)
{
    return data [index >> 3] & (1 << (index & 7));
}

extern (C) bit _darray_set_bit (int value, int index, byte *data)
{
    if (value)
        data [index >> 3] = data [index >> 3] | (1 << (index & 7));
    else
        data [index >> 3] = data [index >> 3] & ~(1 << (index & 7));
    return value;
}

extern (C)
wchar [] _darray_char_to_wchar (char [] array)
{
    wchar [] result;

    result.length = array.length;
    for (int c = 0; c < array.length; c ++)
        result [c] = array [c];

    return result;
}

extern (C)
char [] _darray_wchar_to_char (wchar [] array)
{
    char [] result;
    int d = 0;

    result.length = array.length + 1;
    for (int c = 0; c < array.length; c ++)
        if (array [c] <= char.max)
            result [d ++] = array [c];

    result [d] = \0;
    *(int *) &result = d; /* Stash the length */
    return result;
}

/**********************************************
 * Support for array.reverse property.
 */

extern (C)
Array _adReverse(Array a, int szelem)
    out (result)
    {
        assert(result === a);
    }
    body
    {
        int alen = a.length ();
        
        if (alen >= 2)
        {
            byte *tmp;
            byte [16] buffer;

            void* lo = a.data;
            void* hi = a.data + (alen - 1) * szelem;

            tmp = buffer;
            if (szelem > 16)
                tmp = (byte*) malloc (szelem);

            for (; lo < hi; lo += szelem, hi -= szelem)
            {
                memcpy(tmp, lo,  szelem);
                memcpy(lo,  hi,  szelem);
                memcpy(hi,  tmp, szelem);
            }

            if (szelem > 16)
                free (tmp);
        }
        
        return a;
    }

unittest
{
    debug(adi) printf("array.reverse.unittest\n");

    int[] a = new int[5];
    int[] b;
    int i;

    for (i = 0; i < 5; i++)
        a[i] = i;
    b = a.reverse;
    assert(b === a);
    for (i = 0; i < 5; i++)
        assert(a[i] == 4 - i);

    struct X20
    {   
        // More than 16 bytes in size
        int a;
        int b, c, d, e;
    }

    X20[] c = new X20[5];
    X20[] d;

    for (i = 0; i < 5; i++)
    {   
        c[i].a = i;
        c[i].e = 10;
    }
    d = c.reverse;
    assert(d === c);
    for (i = 0; i < 5; i++)
    {
        assert(c[i].a == 4 - i);
        assert(c[i].e == 10);
    }
}

/**********************************************
 * Support for array.reverse property for bit[].
 */

extern (C) bit [] _adReverseBit (bit [] a)
    out (result)
    {
        assert (result === a);
    }
    body
    {
        if (a.length >= 2)
        {
            bit t;
            int lo, hi;
    
            lo = 0;
            hi = a.length - 1;
            for ( ; lo < hi; lo ++, hi --)
            {
                t = a [lo];
                a [lo] = a [hi];
                a [hi] = t;
            }
        }
        
        return a;
    }

unittest
{
    debug(adi) printf("array.reverse_Bit[].unittest\n");

    bit[] b;
    b = new bit [5];
    bit[] data = [1,0,1,1,0];
    int i;
    
    b[] = data[];
    b.reverse;
    for (i = 0; i < 5; i ++)
        assert (b[i] == data[4 - i]);
}


/**********************************
 * Support for array.dup property.
 */

extern (C) Array _adDup (Array a, int szelem)
    out (result)
    {
        assert (memcmp (result.data, a.data, a.length () * szelem) == 0);
    }
    body
    {
        Array r;
        int size;
        int alen = a.length ();

        size = alen * szelem;
        r.data = (void *) new byte[size];
        r.flength = alen;
        memcpy (r.data, a.data, size);
        return r;
    }

unittest
{
    int[] a;
    int[] b;
    int i;

    debug(adi) printf("array.dup.unittest\n");

    a = new int[3];
    a[0] = 1; a[1] = 2; a[2] = 3;
    b = a.dup;
    assert(b.length == 3);
    for (i = 0; i < 3; i++)
    assert(b[i] == i + 1);
}

unittest
{
    bit [] a;
    
    a = new bit [450];
    assert (a.length == 450);
    assert ((void *) a);
    for (int c = 0; c < a.length; c ++)
        assert (a [c] == 0);
        
    a = new bit [320];
    assert (a.length == 320);
    assert ((void *) a);
    for (int c = 0; c < a.length; c ++)
        assert (a [c] == 0);
}

/**********************************
 * Support for array.dup property for bit[].
 */

extern (C) Array _adDupBit(Array a)
    out (result)
    {
        assert(memcmp(result.data, a.data, (a.length () + 7) / 8) == 0);
    }
    body
    {
        Array r;
        int size;
        int alen = a.length ();

        size = (alen + 31) / 32;
        r.data = (void *) new uint [size];
        r.flength = alen;
        memcpy(r.data, a.data, size * 4);
        return r;
    }

unittest
{
    bit[] a;
    bit[] b;
    int i;

    debug(adi) printf("array.dupBit[].unittest\n");

    a = new bit [3];
    a [0] = 1; a [1] = 0; a [2] = 1;
    b = a.dup;
    assert(b.length == 3);
    for (i = 0; i < 3; i++)
    {
        debug(adi) printf("b[%d] = %d\n", i, b[i]);
        assert(b[i] == (((i ^ 1) & 1) ? true : false));
    }
}


/***************************************
 * Support for array equality test.
 */

extern (C) int _adEq(Array a1, Array a2, TypeInfo ti)
{
    if (a1.length () != a2.length ())
        return 0;       // not equal
    
    int sz = ti.tsize();
    void *p1 = a1.data;
    void *p2 = a2.data;
    int a1len = a1.length ();

    for (int i = 0; i < a1len; i ++)
    {
        if (!ti.equals(p1 + i * sz, p2 + i * sz))
            return 0;       // not equal
    }
    return 1;           // equal
}

unittest
{
    debug(adi) printf("array.Eq unittest\n");

    char[] a = "hello";

    assert(a != "hel");
    assert(a != "helloo");
    assert(a != "betty");
    assert(a == "hello");
}

/***************************************
 * Support for array equality test for bit arrays.
 */

extern (C) int _adEqBit(Array a, Array b)
{
    int alen = a.length (), blen = b.length ();
    int i;

    if (alen != blen)
        return 0;       // not equal
    byte *p1 = cast(byte *) a.data;
    byte *p2 = cast(byte *) b.data;
    uint n = alen / 8;
    
    for (i = 0; i < n; i++)
    {
        if (p1[i] != p2[i])
            return 0;       // not equal
    }

    ubyte mask;

    n = alen & 7;
    mask = (1 << n) - 1;
    //printf("i = %d, n = %d, mask = %x, %x, %x\n", i, n, mask, p1[i], p2[i]);
    return (mask == 0) || (p1 [i] & mask) == (p2 [i] & mask);
}

unittest
{
    debug(adi) printf("array.EqBit unittest\n");

    static bit[] a = [1,0,1,0,1];
    static bit[] b = [1,0,1];
    static bit[] c = [1,0,1,0,1,0,1];
    static bit[] d = [1,0,1,1,1];
    static bit[] e = [1,0,1,0,1];

    assert(a != b);
    assert(a != c);
    assert(a != d);
    assert(a == e);
}

/***************************************
 * Support for array compare test.
 */

extern (C) int _adCmp (Array a, Array b, TypeInfo ti)
{
    int alen = a.length ();
    int blen = b.length ();
    int len;

    //printf("adCmp()\n");
    len = alen;
    if (blen < len)
        len = blen;
    
    int sz = ti.tsize ();
    void *p1 = a.data;
    void *p2 = b.data;
    for (int i = 0; i < len; i++)
    {
        int c;

        c = ti.compare (p1 + i * sz, p2 + i * sz);
        if (c)
            return c;
    }
    
    return cast (int) alen - cast (int) blen;
}

unittest
{
    debug(adi) printf("array.Cmp unittest\n");

    char[] a = "hello";

    assert(a >  "hel");
    assert(a >= "hel");
    assert(a <  "helloo");
    assert(a <= "helloo");
    assert(a >  "betty");
    assert(a >= "betty");
    assert(a == "hello");
    assert(a <= "hello");
    assert(a >= "hello");
}

/***************************************
 * Support for array compare test.
 */

extern (C) int _adCmpChar(Array a, Array b)
{
/+      
version (X86)
{
    asm
    {   naked           ;

        push    EDI     ;
        push    ESI     ;

        mov    ESI,a1+4[4+ESP]  ;
        mov    EDI,a2+4[4+ESP]  ;

        mov    ECX,a1[4+ESP]    ;
        mov    EDX,a2[4+ESP]    ;

    cmp ECX,EDX     ;
    jb  GotLength   ;

    mov ECX,EDX     ;

GotLength:
        cmp    ECX,4        ;
        jb    DoBytes       ;

        // Do alignment if neither is dword aligned
        test    ESI,3       ;
        jz    Aligned       ;

        test    EDI,3       ;
        jz    Aligned       ;
DoAlign:
        mov    AL,[ESI]     ; //align ESI to dword bounds
        mov    DL,[EDI]     ;

        cmp    AL,DL        ;
        jnz    Unequal      ;

        inc    ESI      ;
        inc    EDI      ;

        test    ESI,3       ;

        lea    ECX,[ECX-1]  ;
        jnz    DoAlign      ;
Aligned:
        mov    EAX,ECX      ;

    // do multiple of 4 bytes at a time

        shr    ECX,2        ;
        jz    TryOdd        ;

        repe            ;
    cmpsd           ;

        jnz    UnequalQuad  ;

TryOdd:
        mov    ECX,EAX      ;
DoBytes:
    // if still equal and not end of string, do up to 3 bytes slightly
    // slower.

        and    ECX,3        ;
        jz    Equal     ;

        repe            ;
    cmpsb           ;

        jnz    Unequal      ;
Equal:
        mov    EAX,a1[4+ESP]    ;
        mov    EDX,a2[4+ESP]    ;

        sub    EAX,EDX      ;
        pop    ESI      ;

        pop    EDI      ;
        ret         ;

UnequalQuad:
        mov    EDX,[EDI-4]  ;
        mov    EAX,[ESI-4]  ;

        cmp    AL,DL        ;
        jnz    Unequal      ;

        cmp    AH,DH        ;
        jnz    Unequal      ;

        shr    EAX,16       ;

        shr    EDX,16       ;

        cmp    AL,DL        ;
        jnz    Unequal      ;

        cmp    AH,DH        ;
Unequal:
        sbb    EAX,EAX      ;
        pop    ESI      ;

        or     EAX,1        ;
        pop    EDI      ;

        ret         ;
    }
}
else+/
{
    int alen = a.length ();
    int blen = b.length ();
    int len;
    int c;

//printf("adCmpChar()\n");
    len = alen;
    if (blen < len)
        len = blen;
    c = std.string.memcmp((char *)a.data, (char *)b.data, len);
    if (!c)
    c = cast(int)alen - cast(int)blen;
    return c;
}
}

unittest
{
    debug(adi) printf("array.CmpChar unittest\n");

    char[] a = "hello";

    assert(a >  "hel");
    assert(a >= "hel");
    assert(a <  "helloo");
    assert(a <= "helloo");
    assert(a >  "betty");
    assert(a >= "betty");
    assert(a == "hello");
    assert(a <= "hello");
    assert(a >= "hello");
}

/***************************************
 * Support for array compare test.
 */

extern (C) int _adCmpBit(Array a1, Array a2)
{
    int a1length = a1.length ();
    int a2length = a2.length ();
    int len;
    uint i;

    len = a1length;
    if (a2length < len)
    len = a2length;
    ubyte *p1 = cast(ubyte*)a1.data;
    ubyte *p2 = cast(ubyte*)a2.data;
    uint n = len / 8;
    for (i = 0; i < n; i++)
    {
    if (p1[i] != p2[i])
        break;      // not equal
    }
    for (uint j = i * 8; j < len; j++)
    {   ubyte mask = 1 << j;
    int c;

    c = (int)(p1[i] & mask) - (int)(p2[i] & mask);
    if (c)
        return c;
    }
    return cast(int)a1length - cast(int)a2length;
}

unittest
{
    debug(adi) printf("array.CmpBit unittest\n");

    static bit[] a = [1,0,1,0,1];
    static bit[] b = [1,0,1];
    static bit[] c = [1,0,1,0,1,0,1];
    static bit[] d = [1,0,1,1,1];
    static bit[] e = [1,0,1,0,1];

    assert(a >  b);
    assert(a >= b);
    assert(a <  c);
    assert(a <= c);
    assert(a <  d);
    assert(a <= d);
    assert(a == e);
    assert(a <= e);
    assert(a >= e);
}

