
import internal.object;
import std.string;
import std.c.stdio;

extern (C):

extern (C) void exit (int);

const int PROPMASK = 0x80000000;

struct Array
{
    int flength;
    void *data;
    
    int length () { return flength & 0x7FFFFFFF; }
    void length (int value) { flength = value; }
    
    bit prop () { return flength & 0x80000000; }
    
    void prop (bit value)
    {
        if (!value)
            flength &= 0x7FFFFFFF;
        else
            flength |= 0x80000000;
    }
    
    void safelength (int value)
    {
        if (prop ())
            flength = value & 0x80000000;
        else
            flength = value;
    }
}

int oversize (int size)
{
    uint nbits = 0;
    uint n2 = (uint) size >> 5;
    
    do
    {
        n2 = n2 >> 3;
        nbits += 3;
    }
    while (n2);
        
    return ((size >> nbits) + 1) << nbits;
}

/* Fill the array with a repeated segment */
void _darray_set_repeat (byte [] array, byte [] data)
{
    if (data.length == 1)
        memset (array, data [0], array.length);
    else
    {
        int end = array.length - data.length + 1;
        
        for (int c = 0; c < end; c += data.length)
            array [c .. c + data.length] = data;
    }
}

byte [] _darray_set_length (byte [] x, int length, TypeInfo type)
{
    byte [] init = type.tinit ();
    int size = type.tsize ();
    byte [] a;

    a = new byte [oversize (length) * size];
    if (length < x.length)
        memcpy (a, x, length * size);
    else
    {
        memcpy (a, x, x.length * size);
        _darray_set_repeat (((byte *) a) [x.length * size .. length * size], init);
    }

    *(int *) &a = length; /* Stuff length */
    return a;
}

Array _d_arraycat(Array x, Array y, uint size)
{   
    Array a;
    uint length;
    int xlen = x.length ();
    int ylen = y.length ();
    
    if (!xlen)
        return y;
    if (!ylen)
        return x;

    length = xlen + ylen;
    a.data = new byte [oversize (length) * size];
    a.flength = length;
    memcpy(a.data, x.data, xlen * size);
    //a[0 .. x.length * size] = x[];
    memcpy(&a.data [xlen * size], y.data, ylen * size);
    a.flength = length;
    return a;
}

Array _d_arraycat_safe (Array x, Array y, uint size)
{   
    Array a;
    uint length;
    int xlen = x.length ();
    int ylen = y.length ();
    
    if (!xlen)
        return y;
    if (!ylen)
        return x;

    length = xlen + ylen;
    
    if (x.prop () && oversize (length) == oversize (xlen))
    {
        a.data = x.data;
        a.flength = length;
        memcpy (&a.data [xlen * size], y.data, ylen * size);
        a.prop (true);
    }
    else
    {
        a.data = new byte [oversize (length) * size];
        a.flength = length;
        memcpy (a.data, x.data, xlen * size);
        memcpy (&a.data [xlen * size], y.data, ylen * size);
    }
    
    return a;
}

byte[] _d_arrayappendc(byte[] x, in uint size, ...)
{
    byte[] a;
    uint length;
    void *argp;

    length = x.length + 1;
    a = new byte[length * size];
    memcpy(a, x, x.length * size);
    argp = &size + 1;
    memcpy(&a[x.length * size], argp, size);
    *(int *)&a = length;    // jam length
    x = a;
    return a;
}

Array _d_arrayappendc_safe (Array x, in uint size, ...)
{
    Array a;
    uint length;
    void *argp;
    int xlen = x.length ();

    length = xlen + 1;
    argp = &size + 1;
    if (x.prop () && oversize (length) == oversize (xlen))
    {
        memcpy (&x.data [xlen * size], argp, size);
        x.flength = length | PROPMASK;
        return x;
    }
    else
    {
        a.data = new byte [oversize (length) * size];
        a.flength = length;
        memcpy (a.data, x.data, xlen * size);
        memcpy (&a.data [xlen * size], argp, size);
        a.flength = length | PROPMASK;
        return a;
    }
}

byte[] _d_arraycopy(uint size, byte[] from, byte[] to)
{
    //printf("f = %p,%d, t = %p,%d\n", (void*)from, from.length, (void*)to, to.length);

    if (to.length != from.length)
    {
    throw new Error("lengths don't match for array copy");
    }
    else if ((byte *)to + to.length * size <= (byte *)from ||
    (byte *)from + from.length * size <= (byte *)to)
    {
    memcpy((byte *)to, (byte *)from, to.length * size);
    }
    else
    {
    throw new Error("overlapping array copy");
    }
    return to;
}

bit[] _d_arraycopybit(bit[] from, bit[] to)
{
    //printf("f = %p,%d, t = %p,%d\n", (void*)from, from.length, (void*)to, to.length);
    uint nbytes;

    if (to.length != from.length)
    {
    throw new Error("lengths don't match for array copy");
    }
    else
    {
    nbytes = (to.length + 7) / 8;
    if ((void *)to + nbytes <= (void *)from ||
        (void *)from + nbytes <= (void *)to)
    {
        memcpy((void *)to, (void *)from, nbytes);
    }
    else
    {
        throw new Error("overlapping array copy");
    }
    }
    return to;
}

bit[] _d_arraysetbit(bit[] ba, uint lwr, uint upr, bit value)
in
{
    //printf("_d_arraysetbit(ba.length = %d, lwr = %u, upr = %u, value = %d)\n", ba.length, lwr, upr, value);
    assert(lwr <= upr);
    assert(upr <= ba.length);
}
body
{
    // Inefficient; lots of room for improvement here
    for (uint i = lwr; i < upr; i++)
    ba[i] = value;

    return ba;
}
