
import std.c.stdio;
import std.string;

/******************************************************
 * Support for switch statements switching on strings.
 * Input:
 *  table[]     sorted array of strings generated by compiler
 *  ca      string to look up in table
 * Output:
 *  result      index of match in table[]
 *          -1 if not in table
 */

extern (C):

int _d_switch_string(char[][] table, char[] ca)
    in
    {
    //printf("in _d_switch_string()\n");
    assert(table.length >= 0);
    assert(ca.length >= 0);

    // Make sure table[] is sorted correctly
    int i;

    for (i = 1; i < table.length; i++)
    {
        int len1 = table[i - 1].length;
        int len2 = table[i].length;

        assert(len1 <= len2);
        if (len1 == len2)
        {
        int c;

        c = memcmp(table[i - 1], table[i], len1);
        assert(c < 0);  // c==0 means a duplicate
        }
    }
    }
    out (result)
    {
    int i;
    int c;

    //printf("out _d_switch_string()\n");
    if (result == -1)
    {
        // Not found
        for (i = 0; i < table.length; i++)
        {
        c = memcmp(table[i], ca, ca.length);
        assert(c != 0);
        }
    }
    else
    {
        assert(0 <= result && result < table.length);
        for (i = 0; 1; i++)
        {
        assert(i < table.length);
        c = memcmp(table[i], ca, ca.length);
        if (c == 0)
        {
            assert(i == result);
            break;
        }
        }
    }
    }
    body
    {
    //printf("body _d_switch_string()\n");
    int low;
    int high;
    int mid;
    int c;
    char[] pca;

    low = 0;
    high = table.length;

    // Do binary search
    while (low < high)
    {
        mid = (low + high) >> 1;
        pca = table[mid];
        c = ca.length - pca.length;
        if (c == 0)
        {
        c = memcmp(ca, pca, ca.length);
        if (c == 0)
        {   //printf("found %d\n", mid);
            return mid;
        }
        }
        if (c < 0)
        {
        high = mid;
        }
        else
        {
        low = mid + 1;
        }
    }
    //printf("not found\n");
    return -1;      // not found
    }

int _d_switch_wstring (wchar [] [] table, wchar [] ca)
{
    int low, high, mid, c;
    wchar [] pca;

    low = 0;
    high = table.length;

    while (low < high)
    {
        mid = (low + high) >> 1;
        pca = table [mid];
        c = ca.length - pca.length;
        if (c == 0)
        {
            c = memcmp (ca, pca, ca.length * wchar.size);
            if (c == 0)
                return mid;
        }
        
        if (c < 0)
            high = mid;
        else
            low = mid + 1;
    }

    return -1;
}

int _d_switch_int (int [] table, int ca)
{
    int low, high, mid, c;

    low = 0;
    high = table.length;

    while (low < high)
    {
        mid = (low + high) >> 1;
        c = ca - table [mid];
        if (c < 0)
            high = mid;
        else if (c > 0)
            low = mid + 1;
        else
            return mid;
    }

    return -1;
}

