module std.c.fcntl;

extern (C):
version (Linux)
{
    alias int mode_t;
    alias uint off_t;
    alias uint size_t;
    alias int ssize_t;
    
    /* Bits for the flags parameter of open */
    enum
    {
        O_RDONLY    =      00, /* Open the file for read-only */
        O_WRONLY    =      01, /* Open the file for write-only */
        O_RDWR      =      02, /* Open the file for read/write */
        
        O_CREAT     =    0100, /* If the file doesn't exist it will be created */
        O_EXCL      =    0200, /* When used with O_CREAT, if the file already exists open will fail */
        O_NOCTTY    =    0400, /* If the path refers to a terminal device, it won't become the process' controlling terminal even if the process has none */
        O_TRUNC     =   01000, /* If O_RDWR or O_WRONLY are also specified, truncate the file length to zero */
        O_APPEND    =   02000, /* Before each write, move the file pointer to the end of the file */
        O_NONBLOCK  =   04000, /* Try to open the file in non-blocking mode */
        O_NDELAY    =   04000, /* Equals O_NONBLOCK */
        O_SYNC      =  010000, /* Open for synchronous I/O: any writes block the calling process until it's been physically written */
        O_NOFOLLOW  =  020000, /* Fail if the path is to a symbolic link (Linux 2.1.126 and up only) */
        O_DIRECTORY =  040000, /* Fail if the path is not a directory (Linux 2.1.126 and up only) */
        O_LARGEFILE = 0100000, /* On 32-bit systems which support the Large Files System, allow files whose sizes cannot be represented in 31 bits to be opened */
    }
    
    /* Masks for the mode parameter of open and creat */
    enum
    {
        S_IRWXU = 00700, /* User has read, write, and execute permission */
        S_IRUSR = 00400, /* User has read permission */
        S_IREAD = 00400,
        S_IWUSR = 00200, /* User has write permission */
        S_IWRITE= 00200,
        S_IXUSR = 00100, /* User has execute permission */
        S_IEXEC = 00100,
        
        S_IRWXG = 00070, /* Group has read, write, and execute permission */
        S_IRGRP = 00040, /* Group has read permission */
        S_IWGRP = 00020, /* Group has write permission */
        S_IXGRP = 00010, /* Group has execute permission */
        
        S_IRWXO = 00007, /* Others have read, write, and execute permission */
        S_IROTH = 00004, /* Others have read permission */
        S_IWOTH = 00002, /* Others have write permission */
        S_IXOTH = 00001, /* Others have execute permission */
    }
    
    enum
    {
        SEEK_SET = 0,
        SEEK_CUR = 1,
        SEEK_END = 2,
    }
    
    int close (int fd); 
    int creat (char *pathname, mode_t mode);
    off_t lseek (int fd, off_t offset, int whence);
    int open (char *pathname, int flags);
    int open (char *pathname, int flags, mode_t mode);
    ssize_t read (int fd, void *buf, size_t count);
    ssize_t write (int fd, void *buf, size_t count);
}
