module phobos.test.dbignum;

int _dtest_bignum;

/*

Bignum is largely built up using functions, so it can go wrong by making
incorrect function calls.  So it's important not to assume that a given
function leaves everything proper, even though it has the right result.
The best way to check that is to assert a fmt return with the format
bracketed by integers; if it messes up the stack, it'll cause these to
go haywire.

*/

unittest
{
    bignum x = 45;
    
    assert (x == 45);
    assert (!!x);
    assert (!(bignum) 0);
    assert ((x + x) == 90);
    assert ((x += 2) == 47);
    assert (x == 47);
    assert (x ++ == 47);
    assert (x == 48);
    assert (x -- == 48);
    assert (x == 47);
    assert ((x - 4) == 43);
    assert ((x -= 3) == 44);
    assert (x == 44);
    assert ((x * 640) == 28160);
    assert ((x *= 367) == 16148);
    assert (x == 16148);
    assert ((extended) x == 16148);
    assert ((x / 7) == 2306);
    assert ((x /= 6) == 2691);
    assert (x == 2691);
    assert ((x % 16) == 3);
    assert ((x %= 15) == 6);
    assert (x == 6);
    assert (x < 8);
    assert (x <= 6);
    assert (x > 4);
    assert (x >= 6);
    assert ((x << 2) == 24);
    assert ((x <<= 3) == 48);
    assert (x == 48);
    assert ((x | 13) == 61);
    assert ((x | 56384) == 56432);
//    assert ((x | -364) == -332);
    assert ((x |= 928) == 944);
    assert (x == 944);
    assert ((x & 452) == 384);
    assert ((x &= 675) == 672);
    assert (x == 672);
    assert (~x == -673);
    assert ((x ^ 432) == 784);
    assert ((x ^= 918) == 310);
    assert (x == 310);
    assert (x);
    assert ((x >> 5) == 9);
    assert ((x >>= 2) == 77);
    assert (x == 77);
    print ("%s %s\n", (bignum) 32727.54676, 32727.54676);
    assert ((bignum) 32727.54676 == (bignum) 32727);

    x = 45;
    assert (fmt ("%d %d %d", 4, x == 45,          5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, !!x,              5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, !(bignum) 0,      5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x + x == 90,      5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x += 2) == 47,   5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 47,          5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x ++ == 47,       5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 48,          5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x -- == 48,       5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 47,          5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x - 4 == 43,      5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x -= 3) == 44,   5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 44,          5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x * 640 == 28160, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x *= 367) == 16148, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 16148, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x / 7 == 2306, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x /= 6) == 2691, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 2691, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x % 16 == 3, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x %= 15) == 6, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 6, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x < 8, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x <= 6, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x > 4, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x >= 6, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x << 2 == 24, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x <<= 3) == 48, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 48, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x | 13) == 61, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x | 56384) == 56432, 5) == "4 1 5");
//    assert (fmt ("%d %d %d", 4, (x | -364) == -332, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x |= 928) == 944, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 944, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x & 452) == 384, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x &= 675) == 672, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 672, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, ~x == -673, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x ^ 432) == 784, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x ^= 918) == 310, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 310, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x ? 1 : 0, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x >> 5) == 9, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, (x >>= 2) == 77, 5) == "4 1 5");
    assert (fmt ("%d %d %d", 4, x == 77, 5) == "4 1 5");
}
